@@ -24,9 +24,9 @@ CHANGES:
 As of 0.3.0, options should be specified as `optparse` options.  See README.txt.
 flagReader.py options are still supported for backward compatibility
 """
-import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest, unittest
+import cmd, re, os, sys, optparse, subprocess, tempfile, pyparsing, doctest, unittest, string
 from optparse import make_option
-__version__ = '0.4.2'
+__version__ = '0.4.4'
 
 class OptionParser(optparse.OptionParser):
     def exit(self, status=0, msg=None):
@@ -51,8 +51,8 @@ def options(option_list):
         optionParser.set_usage("%s [options] arg" % func.__name__.strip('do_'))
         def newFunc(instance, arg):
             try:
-                opts, arg = optionParser.parse_args(arg.split())
-                arg = ' '.join(arg)
+                opts, newArgs = optionParser.parse_args(arg.split())
+                newArgs = (newArgs and arg[arg.find(newArgs[0]):]) or ''
             except (optparse.OptionValueError, optparse.BadOptionError,
                     optparse.OptionError, optparse.AmbiguousOptionError,
                     optparse.OptionConflictError), e:
@@ -61,8 +61,9 @@ def options(option_list):
                 return
             if hasattr(opts, '_exit'):
                 return None
+            arg = arg.parser('%s %s%s%s' % (arg.parsed.command, newArgs, arg.parsed.terminator, arg.parsed.suffix))
             result = func(instance, arg, opts)                            
-            return result
+            return result        
         newFunc.__doc__ = '%s\n%s' % (func.__doc__, optionParser.format_help())
         return newFunc
     return option_setup
@@ -137,6 +138,7 @@ else:
         writeToPasteBuffer = getPasteBuffer
           
 pyparsing.ParserElement.setDefaultWhitespaceChars(' \t')
+'''
 def parseSearchResults(pattern, s):
     generator = pattern.scanString(s)
     try:
@@ -147,17 +149,41 @@ def parseSearchResults(pattern, s):
         result = pyparsing.ParseResults('')
         result['before'] = s
     return result
-        
+'''
+
+def replaceInput(source):
+    if source:
+        newinput = open(source[0], 'r').read()
+    else:
+        newinput = getPasteBuffer()
+    
+        try:
+            if statement.inputFrom:
+                newinput = open(statement.inputFrom, 'r').read()
+            else:
+                newinput = getPasteBuffer()
+        except (OSError,), e:
+            print e
+            return 0                    
+        start, end = self.redirectInPattern.scanString(statement.fullStatement).next()[1:]
+        return self.onecmd('%s%s%s' % (statement.fullStatement[:start], 
+                            newinput, statement.fullStatement[end:]))
+
+class ParsedString(str):
+    pass
+
 class Cmd(cmd.Cmd):
     echo = False
     caseInsensitive = True
-    multilineCommands = []
-    continuationPrompt = '> '    
-    shortcuts = {'?': 'help', '!': 'shell', '@': 'load'}
+    continuationPrompt = '> '  
+    legalChars = '!#$%.:?@_' + pyparsing.alphanums + pyparsing.alphas8bit  # make sure your terminators are not in here!
+    shortcuts = {'?': 'help', '!': 'shell', '@': 'load' }
     excludeFromHistory = '''run r list l history hi ed edit li eof'''.split()
     noSpecialParse = 'set ed edit exit'.split()
     defaultExtension = 'txt'
     defaultFileName = 'command.txt'
+    settable = ['prompt', 'continuationPrompt', 'defaultFileName', 'editor', 'caseInsensitive', 'echo']
+    
     editor = os.environ.get('EDITOR')
     _STOP_AND_EXIT = 2
     if not editor:
@@ -168,8 +194,6 @@ class Cmd(cmd.Cmd):
                 if not os.system('which %s' % (editor)):
                     break
             
-    settable = ['prompt', 'continuationPrompt', 'defaultFileName', 'editor', 'caseInsensitive', 'echo']
-    _TO_PASTE_BUFFER = 1
     def do_cmdenvironment(self, args):
         self.stdout.write("""
         Commands are %(casesensitive)scase-sensitive.
@@ -177,7 +201,7 @@ class Cmd(cmd.Cmd):
         Settable parameters: %(settable)s
         """ % 
         { 'casesensitive': ('not ' and self.caseInsensitive) or '',
-          'terminators': self.terminatorPattern,
+          'terminators': str(self.terminators),
           'settable': ' '.join(self.settable)
         })
         
@@ -193,91 +217,223 @@ class Cmd(cmd.Cmd):
     def __init__(self, *args, **kwargs):        
         cmd.Cmd.__init__(self, *args, **kwargs)
         self.history = History()
+        self._init_parser()
         
     def do_shortcuts(self, args):
         """Lists single-key shortcuts available."""
         result = "\n".join('%s: %s' % (sc[0], sc[1]) for sc in self.shortcuts.items())
         self.stdout.write("Single-key shortcuts for other commands:\n%s\n" % (result))
 
-    terminatorPattern = ((pyparsing.Literal(';') ^ pyparsing.Literal('\n\n'))
-                  ^ (pyparsing.Literal('\nEOF') + pyparsing.lineEnd))('terminator')
-    argSeparatorPattern = pyparsing.Word(pyparsing.printables)('command') \
-                          + pyparsing.SkipTo(pyparsing.StringEnd())('args')
-    filenamePattern = pyparsing.Word(pyparsing.alphanums + '#$-_~{},.!:\\/')
-    integerPattern = pyparsing.Word(pyparsing.nums).setParseAction( lambda s,l,t: [ int(t[0]) ] )
-    pipePattern = pyparsing.Literal('|')('pipe') + pyparsing.restOfLine('pipeTo')
-    redirectOutPattern = (pyparsing.Literal('>>') ^ '>')('output') \
-                       + pyparsing.Optional(filenamePattern)('outputTo')
-    redirectInPattern = pyparsing.Literal('<')('input') \
-                      + pyparsing.Optional(filenamePattern)('inputFrom')    
-    punctuationPattern = pipePattern ^ redirectInPattern ^ redirectOutPattern
-    for p in (terminatorPattern, pipePattern, redirectInPattern, redirectOutPattern, punctuationPattern):
-        p.ignore(pyparsing.sglQuotedString)
-        p.ignore(pyparsing.dblQuotedString)    
-
-    def parsed(self, s):
+    commentGrammars = pyparsing.Or([pyparsing.pythonStyleComment, pyparsing.cStyleComment])
+    commentGrammars.addParseAction(lambda x: '')
+    commentInProgress  = pyparsing.Literal('/*') + pyparsing.SkipTo(pyparsing.stringEnd)
+    terminators = [';']
+    blankLinesAllowed = False
+    multilineCommands = []
+    
+    def _init_parser(self):
         '''
         >>> c = Cmd()
-        >>> r = c.parsed('quotes "are > ignored" < inp.txt')
-        >>> r.statement, r.input, r.inputFrom, r.output, r.outputFrom
-        ('quotes "are > ignored" ', '<', 'inp.txt', '', '')
-        >>> r = c.parsed('very complex; < from.txt >> to.txt etc.')
-        >>> r.statement, r.terminator, r.input, r.inputFrom, r.output, r.outputTo
-        ('very complex;', ';', '<', 'from.txt', '>>', 'to.txt')
-        >>> c.parsed('nothing to parse').statement
-        'nothing to parse'
-        >>> r = c.parsed('ignore > within a terminated statement; > out.txt')
-        >>> r.statement, r.terminator, r.input, r.inputFrom, r.output, r.outputTo
-        ('ignore > within a terminated statement;', ';', '', '', '>', 'out.txt')
-        >>> r = c.parsed('send it to | sort | wc')
-        >>> r.statement, r.pipe, r.pipeTo
-        ('send it to ', '|', ' sort | wc')
-        >>> r = c.parsed('got from < thisfile.txt plus blah blah')
-        >>> r.statement, r.input, r.inputFrom
-        ('got from ', '<', 'thisfile.txt')
+        >>> c.multilineCommands = ['multiline']
+        >>> c.caseInsensitive = True
+        >>> c._init_parser()
+        >>> print c.parser.parseString('').dump()        
+        []        
+        >>> print c.parser.parseString('/* empty command */').dump()        
+        []        
+        >>> print c.parser.parseString('plainword').dump()        
+        ['plainword', '']
+        - command: plainword
+        - statement: ['plainword', '']
+          - command: plainword        
+        >>> print c.parser.parseString('termbare;').dump()
+        ['termbare', '', ';', '']
+        - command: termbare
+        - statement: ['termbare', '', ';']
+          - command: termbare
+          - terminator: ;
+        - terminator: ;        
+        >>> print c.parser.parseString('termbare; suffx').dump()
+        ['termbare', '', ';', 'suffx']
+        - command: termbare
+        - statement: ['termbare', '', ';']
+          - command: termbare
+          - terminator: ;
+        - suffix: suffx
+        - terminator: ;        
+        >>> print c.parser.parseString('barecommand').dump()
+        ['barecommand', '']
+        - command: barecommand
+        - statement: ['barecommand', '']
+          - command: barecommand
+        >>> print c.parser.parseString('COMmand with args').dump()
+        ['command', 'with args']
+        - args: with args
+        - command: command
+        - statement: ['command', 'with args']
+          - args: with args
+          - command: command
+        >>> print c.parser.parseString('command with args and terminator; and suffix').dump()
+        ['command', 'with args and terminator', ';', 'and suffix']
+        - args: with args and terminator
+        - command: command
+        - statement: ['command', 'with args and terminator', ';']
+          - args: with args and terminator
+          - command: command
+          - terminator: ;
+        - suffix: and suffix
+        - terminator: ;
+        >>> print c.parser.parseString('simple | piped').dump()
+        ['simple', '', '|', ' piped']
+        - command: simple
+        - pipeTo:  piped
+        - statement: ['simple', '']
+          - command: simple
+        >>> print c.parser.parseString('command with args, terminator;sufx | piped').dump()
+        ['command', 'with args, terminator', ';', 'sufx', '|', ' piped']
+        - args: with args, terminator
+        - command: command
+        - pipeTo:  piped
+        - statement: ['command', 'with args, terminator', ';']
+          - args: with args, terminator
+          - command: command
+          - terminator: ;
+        - suffix: sufx
+        - terminator: ;
+        >>> print c.parser.parseString('output into > afile.txt').dump()
+        ['output', 'into', '>', 'afile.txt']
+        - args: into
+        - command: output
+        - output: >
+        - outputTo: afile.txt
+        - statement: ['output', 'into']
+          - args: into
+          - command: output   
+        >>> print c.parser.parseString('output into;sufx | pipethrume plz > afile.txt').dump()
+        ['output', 'into', ';', 'sufx', '|', ' pipethrume plz', '>', 'afile.txt']
+        - args: into
+        - command: output
+        - output: >
+        - outputTo: afile.txt
+        - pipeTo:  pipethrume plz
+        - statement: ['output', 'into', ';']
+          - args: into
+          - command: output
+          - terminator: ;
+        - suffix: sufx
+        - terminator: ;
+        >>> print c.parser.parseString('output to paste buffer >> ').dump()
+        ['output', 'to paste buffer', '>>', '']
+        - args: to paste buffer
+        - command: output
+        - output: >>
+        - statement: ['output', 'to paste buffer']
+          - args: to paste buffer
+          - command: output
+        >>> print c.parser.parseString('ignore the /* commented | > */ stuff;').dump()
+        ['ignore', 'the /* commented | > */ stuff', ';', '']
+        - args: the /* commented | > */ stuff
+        - command: ignore
+        - statement: ['ignore', 'the /* commented | > */ stuff', ';']
+          - args: the /* commented | > */ stuff
+          - command: ignore
+          - terminator: ;
+        - terminator: ;
+        >>> print c.parser.parseString('has > inside;').dump()
+        ['has', '> inside', ';', '']
+        - args: > inside
+        - command: has
+        - statement: ['has', '> inside', ';']
+          - args: > inside
+          - command: has
+          - terminator: ;
+        - terminator: ;        
+        >>> print c.parser.parseString('multiline has > inside an unfinished command').dump()      
+        ['multiline', ' has > inside an unfinished command']
+        - multilineCommand: multiline        
+        >>> print c.parser.parseString('multiline has > inside;').dump()
+        ['multiline', 'has > inside', ';', '']
+        - args: has > inside
+        - multilineCommand: multiline
+        - statement: ['multiline', 'has > inside', ';']
+          - args: has > inside
+          - multilineCommand: multiline
+          - terminator: ;
+        - terminator: ;        
+        >>> print c.parser.parseString('multiline command /* with comment in progress;').dump()
+        ['multiline', ' command /* with comment in progress;']
+        - multilineCommand: multiline        
+        >>> print c.parser.parseString('multiline command /* with comment complete */ is done;').dump()
+        ['multiline', 'command /* with comment complete */ is done', ';', '']
+        - args: command /* with comment complete */ is done
+        - multilineCommand: multiline
+        - statement: ['multiline', 'command /* with comment complete */ is done', ';']
+          - args: command /* with comment complete */ is done
+          - multilineCommand: multiline
+          - terminator: ;
+        - terminator: ;
         '''
-        if isinstance(s, pyparsing.ParseResults):
-            return s
-        result = (pyparsing.SkipTo(pyparsing.StringEnd()))('fullStatement').parseString(s)
-        command = s.split()[0]
+        outputParser = pyparsing.oneOf(['>>','>'])('output')
+        terminatorParser = pyparsing.Or([(hasattr(t, 'parseString') and t) or pyparsing.Literal(t) for t in self.terminators])('terminator')
+        stringEnd = pyparsing.stringEnd ^ '\nEOF'
+        multilineCommand = pyparsing.Or([pyparsing.Keyword(c, caseless=self.caseInsensitive) for c in self.multilineCommands])('multilineCommand')
+        oneLineCommand = pyparsing.Word(self.legalChars)('command')
+        pipe = pyparsing.Keyword('|', identChars='|')
+        self.commentGrammars.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).setParseAction(lambda x: '')
+        self.commentInProgress.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(pyparsing.cStyleComment)       
+        afterElements = \
+            pyparsing.Optional(pipe + pyparsing.SkipTo(outputParser ^ stringEnd)('pipeTo')) + \
+            pyparsing.Optional(outputParser + pyparsing.SkipTo(stringEnd).setParseAction(lambda x: x[0].strip())('outputTo'))
         if self.caseInsensitive:
-            command = command.lower()
-        '''if command in self.noSpecialParse:
-            result['statement'] = result['upToIncluding'] = result['unterminated'] = result.fullStatement
-            result['command'] = command
-            result['args'] = ' '.join(result.fullStatement.split()[1:])
-            return result'''
-        if s[0] in self.shortcuts:
-            s = self.shortcuts[s[0]] + ' ' + s[1:]
-        result['statement'] = s
-        result['parseable'] = s
-        result += parseSearchResults(self.terminatorPattern, s)
-        if result.terminator:
-            result['statement'] = result.upToIncluding
-            result['unterminated'] = result.before
-            result['parseable'] = result.after
+            multilineCommand.setParseAction(lambda x: x[0].lower())
+            oneLineCommand.setParseAction(lambda x: x[0].lower())
+        if self.blankLinesAllowed:
+            blankLineTerminationParser = pyparsing.NoMatch
         else:
-            result += parseSearchResults(self.punctuationPattern, s)
-            result['statement'] = result['unterminated'] = result.before
-        result += parseSearchResults(self.pipePattern, result.parseable)
-        result += parseSearchResults(self.redirectInPattern, result.parseable)
-        result += parseSearchResults(self.redirectOutPattern, result.parseable)            
-        result += parseSearchResults(self.argSeparatorPattern, result.statement)
-        if self.caseInsensitive:
-            result['command'] = result.command.lower()           
-        result['statement'] = '%s %s' % (result.command, result.args)
-        return result
+            blankLineTerminator = (pyparsing.Literal('\n')('terminator') + stringEnd)
+            blankLineTerminationParser = ((multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(blankLineTerminator).setParseAction(lambda x: x[0].strip())('args') + blankLineTerminator)('statement')
+        multilineParser = (((multilineCommand ^ oneLineCommand) + pyparsing.SkipTo(terminatorParser).setParseAction(lambda x: x[0].strip())('args') + terminatorParser)('statement') +
+             pyparsing.SkipTo(outputParser ^ pipe ^ stringEnd).setParseAction(lambda x: x[0].strip())('suffix') + afterElements)
+        singleLineParser = ((oneLineCommand + pyparsing.SkipTo(terminatorParser ^ stringEnd ^ pipe ^ outputParser).setParseAction(lambda x:x[0].strip())('args'))('statement') +
+            pyparsing.Optional(terminatorParser) + afterElements)
+        self.parser = (
+            stringEnd |
+            multilineParser |
+            multilineCommand + pyparsing.SkipTo(stringEnd) |
+            singleLineParser |
+            blankLineTerminationParser
+            )
+        self.parser.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)
         
-    def extractCommand(self, statement):
-        try:
-            (command, args) = statement.split(None,1)
-        except ValueError:
-            (command, args) = statement, ''
-        if self.caseInsensitive:
-            command = command.lower()
-        return command, args
-       
-    def onecmd(self, line, assumeComplete=False):
+        inputMark = pyparsing.Literal('<')
+        inputMark.setParseAction(lambda x: '')
+        inputFrom = pyparsing.Word(self.legalChars + '/\\')('inputFrom')
+        inputFrom.setParseAction(lambda x: (x and open(x[0]).read()) or getPasteBuffer())
+        self.inputParser = inputMark + pyparsing.Optional(inputFrom)
+        self.inputParser.ignore(pyparsing.sglQuotedString).ignore(pyparsing.dblQuotedString).ignore(self.commentGrammars).ignore(self.commentInProgress)               
+    
+    def parsed(self, raw, **kwargs):
+        if isinstance(raw, ParsedString):
+            p = raw
+        else:
+            s = self.inputParser.transformString(raw.lstrip())
+            for (shortcut, expansion) in self.shortcuts.items():
+                if s.startswith(shortcut):
+                    s = s.replace(shortcut, expansion + ' ', 1)
+                    break
+            result = self.parser.parseString(s)
+            result['command'] = result.multilineCommand or result.command        
+            result['raw'] = raw
+            result['clean'] = self.commentGrammars.transformString(result.args)
+            result['expanded'] = s        
+            p = ParsedString(result.clean)
+            p.parsed = result
+            p.parser = self.parsed
+        for (key, val) in kwargs.items():
+            p.parsed[key] = val
+        return p
+              
+    def onecmd(self, line):
         """Interpret the argument as though it had been typed in response
         to the prompt.
 
@@ -285,65 +441,71 @@ class Cmd(cmd.Cmd):
         see the precmd() and postcmd() methods for useful execution hooks.
         The return value is a flag indicating whether interpretation of
         commands by the interpreter should stop.
+        
+        This (`cmd2`) version of `onecmd` already override's `cmd`'s `onecmd`.
 
         """
-        line = line.strip()
         if not line:
-            return
-        statement = self.parsed(line)
-        while (statement.command in self.multilineCommands) and not \
-              (statement.terminator or assumeComplete):
-            statement = self.parsed('%s\n%s' % (statement.fullStatement, 
-                                    self.pseudo_raw_input(self.continuationPrompt)))
+            return self.emptyline()
+        if not pyparsing.Or(self.commentGrammars).setParseAction(lambda x: '').transformString(line):
+            return 0
+        try:
+            statement = self.parsed(line)
+            while statement.parsed.multilineCommand and (statement.parsed.terminator == ''):
+                statement = self.parsed('%s\n%s' % (statement.parsed.raw, 
+                                        self.pseudo_raw_input(self.continuationPrompt)))
+        except Exception, e:
+            print e
+            return 0
 
+        if not statement.parsed.command:
+            return 0
+        
         statekeeper = None
         stop = 0
-        if statement.input:
-            if statement.inputFrom:
-                try:
-                    newinput = open(statement.inputFrom, 'r').read()
-                except OSError, e:
-                    print e
-                    return 0
-            else:
-                newinput = getPasteBuffer()
-            start, end = self.redirectInPattern.scanString(statement.fullStatement).next()[1:]
-            return self.onecmd('%s%s%s' % (statement.fullStatement[:start], 
-                                newinput, statement.fullStatement[end:]))
-        if statement.pipe and statement.pipeTo:
-            redirect = subprocess.Popen(statement.pipeTo, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
+
+        if statement.parsed.pipeTo:
+            redirect = subprocess.Popen(statement.parsed.pipeTo, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
             statekeeper = Statekeeper(self, ('stdout',))   
             self.stdout = redirect.stdin
-        elif statement.output:
+        elif statement.parsed.output:
             statekeeper = Statekeeper(self, ('stdout',))            
-            if statement.outputTo:
+            if statement.parsed.outputTo:
                 mode = 'w'
-                if statement.output == '>>':
+                if statement.parsed.output == '>>':
                     mode = 'a'
                 try:
-                    self.stdout = open(statement.outputTo, mode)                            
+                    self.stdout = open(statement.parsed.outputTo, mode)                            
                 except OSError, e:
                     print e
                     return 0                    
             else:
                 statekeeper = Statekeeper(self, ('stdout',))
                 self.stdout = tempfile.TemporaryFile()
-                if statement.output == '>>':
+                if statement.parsed.output == '>>':
                     self.stdout.write(getPasteBuffer())
         try:
-            stop = cmd.Cmd.onecmd(self, statement.statement)
+            # "heart" of the command, replace's cmd's onecmd()
+            self.lastcmd = statement.parsed.expanded
+            try:
+                func = getattr(self, 'do_' + statement.parsed.command)
+            except AttributeError:
+                return self.default(statement)
+            stop = func(statement)                        
         except Exception, e:
-            print str(e)
-            return False
+            print e
         try:
-            if statement.command not in self.excludeFromHistory:
-                self.history.append(statement.fullStatement)
+            if statement.parsed.command not in self.excludeFromHistory:
+                self.history.append(statement.parsed.raw)
         finally:
             if statekeeper:
-                if statement.output and not statement.outputTo:
+                if statement.parsed.output and not statement.parsed.outputTo:
                     self.stdout.seek(0)
-                    writeToPasteBuffer(self.stdout.read())
-                elif statement.pipe:
+                    try:
+                        writeToPasteBuffer(self.stdout.read())
+                    except Exception, e:
+                        print str(e)
+                elif statement.parsed.pipeTo:
                     for result in redirect.communicate():              
                         statekeeper.stdout.write(result or '')                        
                 self.stdout.close()
@@ -418,14 +580,8 @@ class Cmd(cmd.Cmd):
         return True
     do_eof = do_EOF
                
-    def clean(self, s):
-        """cleans up a string"""
-        if self.caseInsensitive:
-            return s.strip().lower()
-        return s.strip()
-    
     def showParam(self, param):
-        param = self.clean(param)
+        param = param.strip().lower()
         if param in self.settable:
             val = getattr(self, param)
             self.stdout.write('%s: %s\n' % (param, str(getattr(self, param))))
@@ -447,14 +603,14 @@ class Cmd(cmd.Cmd):
         'Sets a parameter'        
         try:
             paramName, val = arg.split(None, 1)
-            paramName = self.clean(paramName)
+            paramName = paramName.strip().lower()
             if paramName not in self.settable:
                 raise NotSettableError
             currentVal = getattr(self, paramName)
             if (val[0] == val[-1]) and val[0] in ("'", '"'):
                 val = val[1:-1]
             else:                
-                val = cast(currentVal, self.parsed(val).unterminated)
+                val = cast(currentVal, val)
             setattr(self, paramName, val)
             self.stdout.write('%s - was: %s\nnow: %s\n' % (paramName, currentVal, val))
         except (ValueError, AttributeError, NotSettableError), e:
@@ -484,7 +640,7 @@ class Cmd(cmd.Cmd):
                 return self.history.get(arg)[-1]
             else:
                 return self.history[-1]
-        except:
+        except IndexError:
             return None        
     def do_list(self, arg):
         """list [arg]: lists last command issued
@@ -515,15 +671,14 @@ class Cmd(cmd.Cmd):
             print "please use 'set editor' to specify your text editing program of choice."
             return
         filename = self.defaultFileName
-        buffer = ''
-        try:
-            arg = int(arg)
-            buffer = self.last_matching(arg)
-        except:
-            if arg:
+        if arg:
+            try:
+                buffer = self.last_matching(int(arg))
+            except ValueError:
                 filename = arg
-            else:
-                buffer = self.last_matching(arg)
+                buffer = ''
+        else:
+            buffer = self.history[-1]
 
         if buffer:
             f = open(filename, 'w')
@@ -535,7 +690,7 @@ class Cmd(cmd.Cmd):
     do_edit = do_ed
     
     saveparser = (pyparsing.Optional(pyparsing.Word(pyparsing.nums)^'*')("idx") + 
-                  pyparsing.Optional(pyparsing.Word(pyparsing.printables))("fname") +
+                  pyparsing.Optional(pyparsing.Word(legalChars + '/\\'))("fname") +
                   pyparsing.stringEnd)    
     def do_save(self, arg):
         """`save [N] [filename.ext]`
@@ -719,71 +874,53 @@ class OutputTrap(Borg):
     def tearDown(self):
         sys.stdout = self.old_stdout
 
-class TranscriptReader(object):
-    def __init__(self, cmdapp, filename='test_cmd2.txt'):
-        self.cmdapp = cmdapp
-        try:
-            tfile = open(filename)
-            self.transcript = tfile.read()
-            tfile.close()
-        except IOError:
-            self.transcript = ''
-        self.bookmark = 0
-    def refreshCommandFinder(self):
-        prompt = pyparsing.Suppress(pyparsing.lineStart + self.cmdapp.prompt)
-        continuationPrompt = pyparsing.Suppress(pyparsing.lineStart + self.cmdapp.continuationPrompt)
-        self.cmdtxtPattern = (prompt + pyparsing.restOfLine + pyparsing.ZeroOrMore(
-             pyparsing.lineEnd + continuationPrompt + pyparsing.restOfLine))("command")   
-    def inputGenerator(self):
-        while True:
-            self.refreshCommandFinder()
-            (thiscmd, startpos, endpos) = self.cmdtxtPattern.scanString(self.transcript[self.bookmark:], maxMatches=1).next()
-            lineNum = self.transcript.count('\n', 0, self.bookmark+startpos) + 2
-            self.bookmark += endpos
-            yield (''.join(thiscmd.command), lineNum)
-    def nextExpected(self):
-        self.refreshCommandFinder()
-        try:
-            (thiscmd, startpos, endpos) = self.cmdtxtPattern.scanString(self.transcript[self.bookmark:], maxMatches=1).next()
-            result = self.transcript[self.bookmark:self.bookmark+startpos]
-            self.bookmark += startpos
-            return result
-        except StopIteration:
-            return self.transcript[self.bookmark:]
-
 class Cmd2TestCase(unittest.TestCase):
     '''Subclass this, setting CmdApp and transcriptFileName, to make a unittest.TestCase class
        that will execute the commands in transcriptFileName and expect the results shown.
        See example.py'''
-    # problem: this (raw) case gets called by unittest.main - we don't want it to be.  hmm
     CmdApp = None
     transcriptFileName = ''
     def setUp(self):
         if self.CmdApp:
             self.outputTrap = OutputTrap()
             self.cmdapp = self.CmdApp()
-            self.transcriptReader = TranscriptReader(self.cmdapp, self.transcriptFileName)
+            try:
+                tfile = open(self.transcriptFileName)
+                self.transcript = tfile.readlines()
+                tfile.close()
+            except IOError:
+                self.transcript = []
+    def divideTranscript(self):
+        self.dialogue = []
+        commandStart = None
+        response = []
+        command = []
+        prompt = self.cmdapp.prompt.rstrip()
+        continuationPrompt = self.cmdapp.continuationPrompt.rstrip()
+        for (lineNum, line) in enumerate(self.transcript):
+            if line.startswith(prompt):
+                if command:
+                    self.dialogue.append((commandStart, ''.join(command), ''.join(response)))
+                command = [line[len(prompt)+1:]]
+                commandStart = lineNum
+                response = []
+            elif line.startswith(continuationPrompt):
+                command.append(line[len(continuationPrompt)+1:])
+            else:
+                response.append(line)
     def testall(self):
         if self.CmdApp:            
-            for (cmdInput, lineNum) in self.transcriptReader.inputGenerator():
-                self.cmdapp.onecmd(cmdInput, assumeComplete=True)
+            self.divideTranscript()        
+            for (lineNum, command, expected) in self.dialogue:
+                self.cmdapp.onecmd(command)
                 result = self.outputTrap.read()
-                expected = self.transcriptReader.nextExpected()
-                self.assertEqual(self.stripByLine(result), self.stripByLine(expected), 
+                self.assertEqual(result.strip(), expected.strip(), 
                     '\nFile %s, line %d\nCommand was:\n%s\nExpected:\n%s\nGot:\n%s\n' % 
-                    (self.transcriptFileName, lineNum, cmdInput, expected, result))
-    def stripByLine(self, s):
-        bareprompt = self.cmdapp.continuationPrompt.strip()
-        lines = []
-        for line in s.splitlines():
-            line = line.rstrip()
-            if line.startswith(bareprompt):
-                line = line.replace(bareprompt, '', 1)
-            lines.append(line)
-        return '\n'.join(lines).strip()
+                    (self.transcriptFileName, lineNum, command, expected, result))    
     def tearDown(self):
         if self.CmdApp:
             self.outputTrap.tearDown()
         
 if __name__ == '__main__':
-    doctest.testmod()
+    doctest.testmod(optionflags = doctest.NORMALIZE_WHITESPACE)
+    #c = Cmd()
